# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

# -*- coding: utf-8 -*-
"""create_forecast.py

This module contains the CRON job that is periodically executed to make prognoses and
save them in to the database.
This code assumes trained models are available from the persistent storage. If these
are not available run model_train.py to train all models.
To provide the prognoses the folowing steps are carried out:
  1. Get historic training data (TDCV, Load, Weather and APX price data)
  2. Apply features
  3. Load model
  4. Make prediction
  5. Write prediction to the database
  6. Send Teams message if something goes wrong

Example:
    This module is meant to be called directly from a CRON job.
    Alternatively this code can be run directly by running::

        $ python create_forecast.py

Attributes:


"""
from openstf.model import predict
from openstf.model.general import ForecastType
from openstf.tasks.utils.utils import check_status_change, update_status_change
from openstf.tasks.utils.predictionjobloop import PredictionJobLoop
from openstf.tasks.utils.taskcontext import TaskContext


def main():
    with TaskContext(__file__) as context:
        model_type = ["xgb", "xgb_quantile", "lgb"]

        # status file callback after every iteration
        # TODO change implementation to a database one
        def callback(pj, successful):
            status_id = "Pred {}, {}".format(pj["name"], pj["description"])
            status_code = 0 if successful else 2

            if check_status_change(status_id, status_code):
                context.logger.warning("Status changed", status_code=status_code)

                update_status_change(status_id, status_code)

        PredictionJobLoop(
            context,
            model_type=model_type,
            on_end_callback=callback,
            # Debug specific pid
            # prediction_jobs=[{'id':282}],
        ).map(predict.make_prediction, forecast_type=ForecastType.DEMAND)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import datetime

import pandas as pd


def check_max_ahead(pred, max_ahead):
    """Check if the values of pred extend max_ahead hours into the future
    Input:
        - pred: pd.DataFrame(columns=['datetimeFC', *])
        - max_ahead: float, time ahead in hours

    Return:
        - True / False
    """

    # get prediction max ahead
    pred_max_ahead = (
        pd.to_datetime(pred.datetimeFC).max().tz_convert(None)
        - datetime.datetime.utcnow()
    )

    # check max ahead
    if pred_max_ahead.total_seconds() / 60.0 / 60.0 < max_ahead:
        return True
    return False


def check_datetimes_unique(pred):
    """Check if the values of pred only have unique datetimes
    Input:
        - pred: pd.DataFrame(columns=['datetimeFC', *])

    Return:
        - True / False
    """
    # get datetimes
    datetimes = pred.datetimeFC

    # check uniqueness of datetimes
    if datetimes.duplicated().any():
        return True
    return False



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import random
import sys


class PredictionJobLoop:
    def __init__(
        self,
        context,
        stop_on_exception=False,
        random_order=True,
        on_exception_callback=None,
        on_successful_callback=None,
        on_end_callback=None,
        prediction_jobs=None,
        **pj_kwargs
    ):
        """Convenience objects that maps a function over prediction jobs.

        Default behaviour is to automatically get prediction jobs from the
        database. Any keyword argument passed will be directed to the getting
        function. If another set of prediction jobs is desired, manually pass
        them using the prediction_jobs argument.

        Args:
            context (openstf.tasks.util.taskcontext.TaskContext): The
                context to run this loop in.
            stop_on_exception (bool, optional): Whether to break out of the
                loop when an exception is raised. Defaults to False.
            random_order (bool, optional): Whether to randomize the order of the
                prediction jobs. Defaults to True. Does not apply to manually
                passed prediction jobs.
            on_exception_callback (callable, optional): Callback, will be called
                everytime an exception is raised. Callable gets the pj and
                exception raised as arguments
            on_successful_callback (callable, optional): Callback, will be called
                everytime an iteration is successful (no exception is raised).
                Callable gets the pj as argument.
            on_end_callback (callable, optional): Callback, will be called everytime an
                iteration is completed. Callable gets the pj and and bool
                indicating success as argument.
            prediction_jobs (list of dicts, optional): Manually pass a list of
                prediction jobs that will be looped over. A prediction job is a dict
                with at least a value set for the key `id`. If set to None, will fetch
                prediction jobs from database.
            **pj_kwargs: Any other kwargs willed will be directed to the
                prediction job getting function.

        """
        self.context = context
        self.stop_on_exception = stop_on_exception
        self.random_order = random_order
        self.on_exception_callback = on_exception_callback
        self.on_successful_callback = on_successful_callback
        self.on_end_callback = on_end_callback
        self.pj_kwargs = pj_kwargs

        if prediction_jobs is None:
            self.prediction_jobs = self._get_prediction_jobs()
        else:
            self.prediction_jobs = self._add_missing_info_prediction_jobs(
                prediction_jobs
            )

        if self.random_order:
            random.shuffle(self.prediction_jobs)

    def _get_prediction_jobs(self):
        """Fetches prediction jobs from database."""
        self.context.logger.info(
            "Querying prediction jobs from database", **self.pj_kwargs
        )
        prediction_jobs = self.context.database.get_prediction_jobs(**self.pj_kwargs)

        return prediction_jobs

    def _add_missing_info_prediction_jobs(self, prediction_jobs_input):
        prediction_jobs = []

        for pj in prediction_jobs_input:
            if len(pj.keys()) > 1:
                prediction_jobs.append(pj)
                continue

            prediction_jobs.append(self.context.database.get_prediction_job(pj["id"]))

        return prediction_jobs

    def map(self, function, *args, **kwargs):
        """Maps the passed function over all prediction jobs.

        Args:
            function (callable): The function that will be applied to each prediction
                job separately.
            *args: Any other arguments or passed to the function.
            **kwargs: All keyword arguments are passed to the function. This method is
                highly prefered over using args, since kwargs will be automatically
                logged.

        Raises:
            PredictionJobException: This exception will be raised if one or more
            prediction jobs raised an exception during the loop.
        """
        pids_successful = []
        pids_unsuccessful = []
        last_job_exception = None

        num_jobs = len(self.prediction_jobs)
        self.context.perf_meter.checkpoint("pre-loop")

        # loop over prefiction jobs's
        for i, prediction_job in enumerate(self.prediction_jobs):
            successful = False

            self.context.logger = self.context.logger.bind(
                prediction_id=prediction_job.get("id"),
                prediction_name=prediction_job.get("name", ""),
            )

            self.context.perf_meter.start_level(
                "iteration", i, num_jobs=num_jobs, **kwargs
            )

            try:
                function(prediction_job, *args, **kwargs)

                pids_successful.append(prediction_job["id"])
                successful = True

                self._handle_successful_iteration(prediction_job)

            except Exception as exception:
                pids_unsuccessful.append(prediction_job["id"])
                last_job_exception = exception

                self._handle_exception_during_iteration(prediction_job, exception)

                if self.stop_on_exception:
                    break
            finally:
                self._handle_finished_last_iteration(prediction_job, successful)

            self.context.logger = self.context.logger.unbind("prediction_id")

        jobs_successful = len(pids_successful)
        jobs_unsuccessful = len(pids_unsuccessful)
        jobs_started = jobs_successful + jobs_unsuccessful

        # This log is for human readable logging
        self.context.perf_meter.checkpoint(
            "loop",
            num_jobs=num_jobs,
            jobs_started=jobs_started,
            jobs_successful=jobs_successful,
            jobs_unsuccessful=jobs_unsuccessful,
            successful=int(jobs_unsuccessful > 0),
        )

        if jobs_unsuccessful > 0:
            metrics = {
                "num_jobs": num_jobs,
                "pids_successful": pids_successful,
                "pids_unsuccessful": pids_unsuccessful,
                "jobs_successful": jobs_successful,
                "jobs_unsuccessful": jobs_unsuccessful,
                "jobs_started": jobs_started,
            }

            raise PredictionJobException(metrics) from last_job_exception

    def _handle_successful_iteration(self, prediction_job):
        if self.on_successful_callback is not None:
            try:
                self.on_successful_callback(prediction_job)
            except Exception:
                _, exc_info, stack_info = sys.exc_info()
                self.context.logger.error(
                    "An exception occured when executing the on_successful_callback\
                        callback function for this iteration",
                    exc_info=exc_info,
                    stack_info=stack_info,
                )

    def _handle_exception_during_iteration(self, prediction_job, e):
        _, exc_info, stack_info = sys.exc_info()
        self.context.logger.error(
            "An exception occured during this iteration",
            exc_info=exc_info,
            stack_info=stack_info,
        )

        if self.on_exception_callback is None:
            return

        try:
            self.on_exception_callback(prediction_job, e)
        except Exception:
            _, exc_info, stack_info = sys.exc_info()
            self.context.logger.error(
                "An exception occured when executing the on_exception_callback "
                "callback function for this iteration",
                exc_info=exc_info,
                stack_info=stack_info,
            )

    def _handle_finished_last_iteration(self, prediction_job, successful):

        self.context.perf_meter.complete_level(successful)

        if self.on_end_callback is not None:
            try:
                self.on_end_callback(prediction_job, successful)
            except Exception:
                _, exc_info, stack_info = sys.exc_info()
                self.context.logger.error(
                    "An exception occured when executing the on_end_callback callback\
                    function for this iteration",
                    exc_info=exc_info,
                    stack_info=stack_info,
                )


class PredictionJobException(Exception):
    def __init__(self, metrics={}):
        super(PredictionJobException, self).__init__(
            "One or more prediction jobs raised an exception."
        )
        self.metrics = metrics



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

"""create_wind_forecast
This module contains the CRON job that is periodically executed to make prognoses of
wind features.
These features are usefull for splitting the load in solar and wind contributions and
making prognoses.
Example:
    This module is meant to be called directly from a CRON job. A description of the
    CRON job can be found in the /k8s/CronJobs folder.
    Alternatively this code can be run directly by running::
        $ python create_wind_forecast
Attributes:

"""

from openstf.feature_engineering import apply_features
from openstf.tasks.utils.predictionjobloop import PredictionJobLoop
from openstf.tasks.utils.taskcontext import TaskContext


def make_wind_forcast_pj(pj, context):
    """Make a wind prediction for a specific prediction job

    Args:
        pj: (dict) prediction job
    """
    context.logger.info("Get turbine data", turbine_type=pj["turbine_type"])
    turbine_data = context.database.get_power_curve(pj["turbine_type"])

    context.logger.info(
        "Get windspeed", location=[pj["lat"], pj["lon"]], hub_height=pj["hub_height"]
    )
    windspeed = context.database.get_wind_input(
        (pj["lat"], pj["lon"]),
        pj["hub_height"],
        pj["horizon_minutes"],
        pj["resolution_minutes"],
    )

    context.logger.info("Calculate windturbine power", n_turbines=pj["n_turbines"])
    power = apply_features.calculate_windturbine_power_output(
        windspeed, pj["n_turbines"], turbine_data
    ).rename(columns=dict(windspeed_100m="forecast"))

    context.logger.info("Store wind prediction in database")
    power["pid"] = pj["id"]
    power["type"] = "wind"
    power["algtype"] = "powerCurve"
    power["customer"] = pj["name"]
    power["description"] = pj["description"]
    context.database.write_forecast(power, t_ahead_series=True)


def main():
    with TaskContext(__file__) as context:
        context.logger.info("Querying wind prediction jobs from database")
        prediction_jobs = context.database.get_prediction_jobs_wind()
        prediction_jobs = [x for x in prediction_jobs if x["model"] == "latest"]

        PredictionJobLoop(context, prediction_jobs=prediction_jobs).map(
            make_wind_forcast_pj, context
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import pandas as pd
from ktpbase.database import DataBase
from ktpbase.log import logging

from openstf.feature_engineering.capacity_prognoses_features import (
    apply_capacity_features,
)
from openstf.model.capacity.model import CapacityPredictionModel
from openstf.model.capacity.utils import prepare_training_data
from openstf.model.serializer.xgboost.xgboost import XGBModelSerializer


def train_capacity_prognosis(pj, datetime_start, datetime_end, y_hor=[0, 6, 13]):
    """Train a capacity prognoses model for a specific prediction job.

    Args:
        pj: (dict) prediction job
    """
    db = DataBase()

    # initialize logging
    logger = logging.get_logger(__name__)

    # get input data
    logger.info("Get input data")
    load_data = db.get_load_pid(
        pid=pj["id"], datetime_start=datetime_start, datetime_end=datetime_end
    )
    load_profiles_data = db.get_tdcv_load_profiles(datetime_start, datetime_end)
    load_profile_names = list(load_profiles_data.columns)

    input_data = pd.concat([load_data, load_profiles_data], axis=1)

    # apply features
    logger.info("Apply features")
    feature_data, classes = apply_capacity_features(
        input_data,
        y_col="load_max",
        y_hor=y_hor,
        outlier_removal=False,
        load_profile_names=load_profile_names,
    )

    # prepare data
    logger.info("Prepare prediction data")
    train_x, train_y, train_h, val_x, val_y, val_h = prepare_training_data(
        feature_data, y_col="load_max"
    )

    # create model
    logger.info("Create capacity prognosis model")
    model = CapacityPredictionModel(classes=classes)

    # train model
    logger.info("Train model")
    model.train(train_x, train_y, val_x, val_y)

    # evaluate model performance
    logger.info("Evaluate model")
    scores = {
        "train": model.evaluate(train_x, train_y),
        "validation": model.evaluate(val_x, val_y),
    }  # Gives code smell but is necessary for unit tests
    # TODO: make train report?
    logger.info("Training metrics", scores=scores)

    # save model
    logger.info("Saving model")
    serializer = XGBModelSerializer()
    directory = serializer.trained_models_folder / f'{pj["id"]}' / "capacity"
    model.save(directory=directory)



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import dill
from ktpbase.log import logging


class XGBQuantileModel:
    def __init__(self, quantile_models=None, model_file=None):

        self.logger = logging.get_logger(self.__class__.__name__)

        if quantile_models is not None and model_file is None:
            self.quantile_models = quantile_models
            self.best_ntree_limit = self.quantile_models[0.5].best_ntree_limit
            self.feature_names = self.quantile_models[0.5].feature_names
            self.feature_types = self.quantile_models[0.5].feature_types
        elif model_file is not None and quantile_models is None:
            self._load_from_disk(model_file)
        else:
            # TODO describe reason and cases (if en elif case)
            print("Could not initialize!")

        # Check if 0.5 quantile is present, raise exception otherwise
        if 0.5 not in self.quantile_models.keys():
            raise ValueError("Quantile model does not have a model for the median")

    def predict(self, *args, quantile=0.5, **kwargs):
        return self.quantile_models[quantile].predict(*args, **kwargs)

    def get_score(self, *args, quantile=0.5, **kwargs):
        return self.quantile_models[quantile].get_score(*args, **kwargs)

    def save_model(self, save_location):

        # Save dict with quantile models
        save_location = save_location.parent / "model_quantile.bin"
        dill.dump(self.quantile_models, file=open(save_location, "wb"))

        # Save best n_trees
        for quantile in self.quantile_models.keys():
            try:
                save_location = save_location.parent / f"best_iteration_{quantile}.pkl"
                dill.dump(
                    self.quantile_models[quantile].best_ntree_limit,
                    file=open(save_location, "wb"),
                )
            except Exception as e:
                print(
                    "No best iteration found in the output, best iteration not saved for quantile "
                    + str(quantile)
                    + ".",
                    e,
                )

    def _load_from_disk(self, model_file):
        # Save dict with quantile models
        self.quantile_models = dill.load(open(model_file, "rb"))
        # Save best n_trees
        for quantile in self.quantile_models.keys():
            try:
                filename = model_file.parent / f"best_iteration_{quantile}.pkl"
                self.quantile_models[quantile].best_ntree_limit = dill.load(
                    open(filename, "rb")
                )
            except Exception as e:
                print(
                    "No best iteration found for quantile " + str(quantile) + ".",
                    e,
                )
                self.quantile_models[quantile].best_ntree_limit = 15



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from functools import partial

import xgboost as xgb
from ktpbase.log import logging

from openstf.model import metrics
from openstf.model.prediction.xgboost.model.quantile import XGBQuantileModel
from openstf.model.trainer.xgboost.xgboost import XGBModelTrainer

# Available trainings period durations for optimization
# After preprocessing, the data consists of 75 days (of the original 90 days).
# To prevent overfitting on a very short period of time, the test and validation
# sets are always a constant fraction of the total data period (75 days).
# We take an optimized fraction of the remaining data as our train-data.
# Given the current fractions for test and validation, the longest train-block
# is 60 days.


class XGBQuantileModelTrainer(XGBModelTrainer):
    def __init__(self, pj):
        super().__init__(pj)
        self.logger = logging.get_logger(self.__class__.__name__)

    def train(
        self, train_data, validation_data, callbacks=None, early_stopping_rounds=10
    ):
        """

        Returns:
            (dict): with keys 'quantile_PXX' and the trained models as values
        """

        # Convert train and validation sets to Dmatrix format for computational
        #  efficiency. Drop Horizon column
        dtrain = xgb.DMatrix(
            train_data.iloc[:, 1:].drop("Horizon", axis=1, errors="ignore"),
            label=train_data.iloc[:, 0],
        )  # [:,1:-1] excludes label and 'horizon' column
        dval = xgb.DMatrix(
            validation_data.iloc[:, 1:].drop("Horizon", axis=1, errors="ignore"),
            label=validation_data.iloc[:, 0],
        )

        # Define data set to be monitored during training, the last(validation)
        #  will be used for early stopping
        watchlist = [(dtrain, "train"), (dval, "validation")]

        # Create result dictionary to house the models for the different quantiles
        self.quantile_models = dict()

        # get the xgb (hyper) parameters
        params = {k: self.hyper_parameters[k] for k in self._xgb_hyper_parameter_keys}

        for quantile in self.pj["quantiles"]:
            self.logger.info(f"Training quantile '{quantile}' model", quantile=quantile)

            # Define objective callback functions specifically for desired quantile
            xgb_quantile_eval_this_quantile = partial(
                metrics.xgb_quantile_eval, quantile=quantile
            )
            xgb_quantile_obj_this_quantile = partial(
                metrics.xgb_quantile_obj, quantile=quantile
            )

            # Train quantile model
            self.quantile_models[quantile] = xgb.train(
                params=params,
                dtrain=dtrain,
                evals=watchlist,
                num_boost_round=500,  # Can be large because we are early stopping anyway
                obj=xgb_quantile_obj_this_quantile,
                feval=xgb_quantile_eval_this_quantile,
                verbose_eval=False,
                early_stopping_rounds=early_stopping_rounds,
                callbacks=callbacks,
            )

        self.trained_model = XGBQuantileModel(self.quantile_models)
        return self.trained_model



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

"""run_tracy.py
Tracy checks the mysql todolist and tries her best to execute the
functions with desired inputs
This scripts works as follows:
  1. Checks the mysql table 'todolist' for jobs (which are not already in progress and
    which are not already failed)
  2. Set all newly aquired jobs to 'in progress'
For each job;
  3. Convert input arguments to a dict with 'args' and 'kwargs'
  4. Interpret the given function and arguments
  5. Execute the job
  6. Post result to Slack
  7. Remove job from mysql table
If job fails, set in progress to 2
All functions that tracy is able to execute need to be imported and defined in the
available_functions.

Example:
    This module is meant to be called directly from a CRON job.

    Alternatively this code can be run directly by running::
        $ python run_tracy.py
Attributes:
"""

# sql om de todolist tabel te maken:

# CREATE TABLE IF NOT EXISTS `tst_icarus`.`todolist` (
# `id` INT NOT NULL AUTO_INCREMENT ,
# `created` DATETIME NOT NULL DEFAULT CURRENT_TIMESTAMP ,
# `function` VARCHAR(200) NOT NULL ,
# `args` VARCHAR(200) NOT NULL ,
# `inprogress` BOOLEAN NULL DEFAULT NULL ,
# PRIMARY KEY (`id`), UNIQUE `id` (`id`))
# ENGINE = InnoDB;

from functools import partial

from openstf.model.hyper_parameters import optimize_hyperparameters
from openstf.model.train import train_specific_model

# Import project modules
from openstf.monitoring.teams import post_teams
from openstf.tasks.utils.utils import (
    convert_string_args_to_dict_args,
    interpret_string_as_functions,
)
from openstf.tasks.utils.taskcontext import TaskContext


def get_and_evaluate_todos(context):
    # Get the to do list.
    todo_dict = context.database.ktp_api.get_all_tracy_jobs(inprogress=0)
    num_jobs = len(todo_dict)

    if num_jobs == 0:
        context.logger.warning("No TODO jobs to perform, exit", num_jobs=num_jobs)
        return

    context.logger.info("Start performing TODO jobs", num_jobs=num_jobs)

    for i, job in enumerate(todo_dict):
        # Set all retrieved items of the todolist to inprogress
        job["inprogress"] = 1
        context.database.ktp_api.update_tracy_job(job)

        context.logger = context.logger.bind(job_id=job["id"])
        context.logger.info("Process job", job_counter=i, total_jobs=num_jobs)

        # First try to parse the input directly
        try:
            interpreted_func = job["function"] + "(" + job["args"] + ")"
        except Exception as e:
            context.logger.error(
                "An exception occured while interpreting function", exc_info=e, job=job
            )
            context.logger.info("Try alternative parsing of function and arguments")

            function_arguments = convert_string_args_to_dict_args(job["args"])
            interpreted_func = interpret_string_as_functions(
                job["function"],
                function_arguments["args"],
                function_arguments["kwargs"],
            )

        context.logger.info("Interpreted the job", interpreted_func=interpreted_func)

        # Make eval safe by exluding all global functions, except the functions from func
        # Do this by setting al usually available global functions to None,
        # And adding func explicitely

        available_functions = {
            "train_specific_model": partial(train_specific_model, context),
            "optimize_hyperparameters_for_specific_pid": optimize_hyperparameters,
        }

        forbid_globals = {key: None for key in globals().keys()}
        forbid_globals.update(available_functions)

        # Try to evaluate the function
        try:
            # TODO eval is considered bad practice in (almost) every case
            res = eval(interpreted_func, forbid_globals)
            context.logger.debug(
                "Evaluated interpreted function",
                return_value=res,
                interpreted_func=interpreted_func,
            )

            teams_message = (
                f"Tracy executed a job. Job: `{interpreted_func}`   \n"
                f"Result: {str(res)}"
            )
            # TODO is this not redundant? for example optimize_hyperparameters
            # already sends a teams message with the results?
            post_teams(teams_message)

            # Remove job from mysql todolist
            context.database.ktp_api.delete_tracy_job(job)
            context.logger.info("Job removed from mysql table")
        except Exception as e:
            job["inprogress"] = 2
            context.database.ktp_api.update_tracy_job(job)
            context.logger.error(
                "An exception occured during evaluation of the interpreted function",
                exc_info=e,
            )
            teams_message = (
                f"Tracy could not execute job (`{interpreted_func}`), "
                f"set inprogress to 2"
            )
            post_teams(teams_message)

    context.logger = context.logger.unbind("job_id")
    context.logger.info("Finished all Tracy jobs - Tracy out!")


def main():
    with TaskContext(__file__) as context:
        get_and_evaluate_todos(context)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

# -*- coding: utf-8 -*-
"""create_components_forecast.py

This module contains the CRON job that is periodically executed to make
the components prognoses and save them in the database.

This code assumes trained models are available from the persistent storage.
If these are not available run model_train.py to train all models.
To provide the prognoses the following steps are carried out:
  1. Get historic training data (TDCV, Load, Weather and APX price data)
  2. Apply features
  3. Load model
  4. Make component prediction
  5. Write prediction to the database
  6. Send Teams message if something goes wrong

Example:
    This module is meant to be called directly from a CRON job. A description of
    the CRON job can be found in the /k8s/CronJobs folder.
    Alternatively this code can be run directly by running::

        $ python create_components_forecast.py

Attributes:


"""
import openstf.model.predict as predict
from openstf.tasks.utils.utils import check_status_change, update_status_change
from openstf.tasks.utils.predictionjobloop import PredictionJobLoop
from openstf.tasks.utils.taskcontext import TaskContext


def create_components_forecast_pj(pj, context):
    if pj["train_components"] == 0:
        context.logger.info(
            "Skip prediction job", train_components=pj["train_components"]
        )
        return

    # Make forecast for the demand, wind and pv components
    predict.make_components_prediction(pj)


def main():
    with TaskContext(__file__) as context:

        # status file callback after every iteration
        # TODO change implementation to a database one
        def callback(pj, successful):
            status_id = "Pred {}, {}".format(pj["name"], pj["description"])
            status_code = 0 if successful else 2

            if check_status_change(status_id, status_code):
                context.logger.warning("Status changed", status_code=status_code)

                update_status_change(status_id, status_code)

        model_type = ["xgb", "xgb_quantile", "lgb"]

        PredictionJobLoop(
            context,
            model_type=model_type,
            on_end_callback=callback,
        ).map(create_components_forecast_pj, context)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

# -*- coding: utf-8 -*-
"""run_validationrobot.py

This module contains the CRON job that is periodically executed to detect
non zero stationsflatliners.

To provide the detection the folowing steps are caried out:
  1. Get historic training data (TDCV, Load, Weather and APX price data)
  2. Set threshold
  3. Detect flatliners
  4. Send slack message if at least one flatliner is found

Example:
    This module is meant to be called directly from a CRON job.

    Alternatively this code can be run directly by running::

        $ python run_validationrobot.py

Attributes:


"""
# Import builtins
from datetime import datetime, timedelta
from pathlib import Path

import cufflinks
import pandas as pd

# Import third-party
import plotly

# Import project modules
from openstf.model.validation_robot import nonzero_flatliner, zero_flatliner
from openstf.monitoring.teams import post_teams
from openstf.tasks.utils.predictionjobloop import PredictionJobLoop
from openstf.tasks.utils.taskcontext import TaskContext

cufflinks.go_offline()
colors = {"Flatliner": "rgba(255, 0, 0, 0.3)", "Zero_value": "rgba(0, 255, 0, 0.3)"}


def validation_robot_pj(pj, context, datetime_start, datetime_end):
    """Run validation robot for specific prediction job.

    Args:
        pj: (dict) prediction job
        datetime_start: (datetime.datetime) start date
        datetime_end: (datetime.datetime) end date

    Returns:
        None

    """
    # Get load data
    load_data = context.database.get_load_pid(
        pj["id"], datetime_start, datetime_end, aggregated=False
    )
    if len(load_data) == 0:
        context.logger.warning("No load data found, skipping")
        return

    # Detect flatliners and zero-values
    load_errors = pd.DataFrame()

    flatliner_moments = nonzero_flatliner(load_data, threshold=4)
    zero_value_moments = zero_flatliner(load_data, threshold=4)

    # Combine single errors to singel result dataframe
    # Act if flatliner occurs
    if flatliner_moments is not None:
        num_flatliners = len(flatliner_moments)
        context.logger.warning(
            "Flatliner detected",
            prediction_name=pj["name"],
            num_flatliners=num_flatliners,
        )

        flatliner_moments["type"] = "Flatliner"
        load_errors = load_errors.append(flatliner_moments)
    # Same for zero-values
    if zero_value_moments is not None:
        num_zero_values = len(zero_value_moments)
        context.logger.warning(
            "Zero-value detected",
            prediction_name=pj["name"],
            num_zero_values=num_zero_values,
        )
        zero_value_moments["type"] = "Zero_value"
        load_errors = load_errors.append(zero_value_moments)

    num_load_errors = len(load_errors)

    # If some error was found, proceed to make a nice image and send
    # a message to Teams
    if num_load_errors > 0:
        context.logger.warning(
            "Total load errors detected", num_load_errors=num_load_errors
        )
        # Make nice image
        plot_url = _create_load_error_plot(pj, load_data, load_errors, context.config)
        context.logger.info("Created a load error plot", url=plot_url)
        # Send teams error message
        _send_teams_error_message(pj, num_load_errors, plot_url)


def _create_load_error_plot(pj, load_data, load_errors, config):
    start = load_errors["from_time"]
    end = load_errors["to_time"]
    check_type = load_errors["type"]
    vspanlist = [
        {
            "x0": from_time,
            "x1": to_time,
            "color": colors[check_type],
            "fill": True,
            "opacity": 0.4,
        }
        for from_time, to_time, check_type in zip(start, end, check_type)
    ]
    fig = load_data.iplot(vspan=vspanlist, asFigure=True)

    # Define image save location
    # NOTE should this also be moved to the config?
    image_name = "stations_detection.html"

    base_path = config.paths.trained_models
    base_url = config.dashboard.trained_models_url

    filename = Path(f'{base_path}/{pj["id"]}/{image_name}')

    web_link = f'{base_url}/{pj["id"]}/{image_name}'

    plotly.offline.plot(
        fig,
        show_link=False,
        filename=str(filename),
        auto_open=False,
    )

    return web_link


def _send_teams_error_message(pj, num_load_errors, buttonurl):
    # Prepare  message
    msg = {
        "fallback": "Load data error(s) detected {0}: {1} timeperiods".format(
            pj["name"], num_load_errors
        ),
        "title": "Load data error(s) detected {0} {1}".format(pj["name"], pj["id"]),
        "links": dict(buttontext="Robot Result", buttonurl=buttonurl),
        "text": (
            "Found {0} moments where there is a load error at pid {1}!".format(
                num_load_errors, pj["id"]
            )
        ),
        "color": "#764FA5",
    }
    post_teams(msg)


def main():
    with TaskContext(__file__) as context:
        datetime_start = datetime.utcnow() - timedelta(days=7)
        datetime_end = datetime.utcnow()

        PredictionJobLoop(context).map(
            validation_robot_pj,
            context,
            datetime_start=datetime_start,
            datetime_end=datetime_end,
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

# -*- coding: utf-8 -*-
from datetime import timedelta

import numpy as np
import pandas as pd


def nonzero_flatliner(df, threshold):
    """Function that detects a stationflatliner and returns a list of datetimes.

    Args:
        df: pd.dataFrame(index=DatetimeIndex, columns = [load1, ..., loadN]).
            Load_corrections should be indicated by 'LC_'
        threshold: after how many timesteps should the function detect a flatliner.

    Returns:
    # TODO: function returns None or a DataFrame
        list: flatline moments
    """

    if len(df) == 0:
        return

    # remove load corrections
    df = df.loc[:, ~df.columns.str.startswith("LC_")]

    # Remove moments when total load is 0
    df = df[(df.T != 0).any()]

    # Add columns with diff
    diff_columns = ["diff_" + x for x in df.columns]
    df[diff_columns] = df.diff().abs()
    # We are looking for when total station is a flatliner, so all
    # Select where all diff columns are zero
    flatliner_total = (df[diff_columns] == 0).mean(axis=1) == 1
    # Find all first and last occurences of 0
    first_zeros = flatliner_total[:-1][
        (flatliner_total.iloc[:-1] == 0).values * (flatliner_total.iloc[1:] != 0).values
    ]
    last_zeros = flatliner_total[1:][
        (flatliner_total.iloc[:-1] != 0).values * (flatliner_total.iloc[1:] == 0).values
    ]
    last_zeros.index = last_zeros.index - pd.Timedelta("15m")

    # Give as output from:to
    interval_df = pd.DataFrame(
        {
            "from_time": first_zeros.index,
            "to_time": last_zeros.index,
            "duration_h": last_zeros.index - first_zeros.index,
        },
        index=range(len(first_zeros)),
    )

    # Only keep periods which exceed threshold
    interval_df = interval_df[interval_df["duration_h"] >= timedelta(hours=threshold)]
    if len(interval_df) == 0:
        interval_df = None
    return interval_df


def check_data_for_each_trafo(df, col):
    """Function that detects if each column contains zero-values at all, only zero-values and NaN values.

    Args:
        df: pd.dataFrame(index=DatetimeIndex, columns = [load1, ..., loadN]).
            Load_corrections should be indicated by 'LC_'
        col: column of pd.dataFrame

    Returns:
        bool: False if column contains above specified or True if not"""
    if df is not None:
        # Check for each column the data on the following: (Skipping if true)
        # Check if there a zero-values at all
        if (df[col] != 0).all(axis=0):
            print(
                "No zero values found - at all at trafo {a}, skipping column".format(
                    a=col
                )
            )
            return False
        # Check if all values are zero in column
        elif (df[col] == 0).all(axis=0):
            print("Load at trafo {a} is zero, skipping column".format(a=col))
            return False
        # Check if all values are NaN in the column
        elif np.all(pd.isna(col)):
            print("Load at trafo {a} is missing, skipping column".format(a=col))
            return False
        return True
    else:
        return False


def zero_flatliner(df, threshold, window=timedelta(minutes=30), load_threshold=0.3):
    """Function that detects a zero value where the load is not compensated by the other trafo's of the station.

    Input:
    - df: pd.dataFrame(index=DatetimeIndex, columns = [load1, ..., loadN]). Load_corrections should be indicated by 'LC_'
    - threshold (float): after how many hours should the function detect a flatliner.
    - window (timedelta object): for how many hours before the zero-value should the mean load be calculated.
    - load_threshold (fraction): how big may the difference be between the total station load
    before and during the zero-value(s).

    return:
    - pd.DataFrame of timestamps, or None if none"""
    result_df = pd.DataFrame()

    for col in df.columns:
        # Check for each trafo if it contains zero-values or NaN values
        if check_data_for_each_trafo(df, col):
            # Copy column in order to manipulate data
            df_col = df[col].copy()
            # Create list of zero_values to check length
            zero_values = df_col == 0
            last_zeros = zero_values[:-1][
                (zero_values.iloc[:-1] != 0).values * (zero_values.iloc[1:] == 0).values
            ]
            first_zeros = zero_values[1:][
                (zero_values.iloc[:-1] == 0).values * (zero_values.iloc[1:] != 0).values
            ]
            interval_df = pd.DataFrame(
                {
                    "from_time": first_zeros.index,
                    "to_time": last_zeros.index,
                    "duration_h": last_zeros.index - first_zeros.index,
                },
                index=range(len(first_zeros)),
            )

            # Keep dataframe with zero_value > threshold
            interval_df = interval_df[
                interval_df["duration_h"] >= timedelta(hours=threshold)
            ]

            # Check if there are missing values in the data
            num_nan = np.sum(pd.isna(df_col))
            if np.sum(num_nan) > 0:
                print(
                    "Found {a} missing values at trafo {b}, dropping NaN values".format(
                        a=num_nan, b=col
                    )
                )
                df_col = df_col.dropna()

            non_compensated_df = pd.DataFrame()
            print(
                "Found {b} suspicious moments were load is a zero at trafo {a}".format(
                    a=col, b=len(interval_df)
                )
            )
            # Forloop all moments of zero_values, check if delta of to_zero_trafo is compesated by other trafos
            for i, candidate in interval_df.iterrows():
                # Calculate mean load before and after zero-moment
                mean_load_before = (
                    df[candidate.from_time - window : candidate.from_time]
                    .sum(axis=1)
                    .mean()
                )
                mean_full_timespan = (
                    df[candidate.from_time : candidate.from_time + window]
                    .sum(axis=1)
                    .mean()
                )

                # Compare load and detect zero-value flatliner
                if (
                    np.abs(
                        (mean_load_before - mean_full_timespan)
                        / np.max([np.abs(mean_load_before), np.abs(mean_full_timespan)])
                    )
                    > load_threshold
                ):
                    non_compensated_df = non_compensated_df.append(candidate)
                    print(
                        "Found a non-compensated zero value:",
                        candidate.to_string(index=False),
                    )

            # after checking all candidates for a single trafo, add moments to result_df
            result_df = result_df.append(non_compensated_df)

    if len(result_df) == 0:
        result_df = None
    return result_df


def replace_invalid_data(df, suspicious_moments):
    """Function that detects invalid data using the nonzero_flatliner function and converts the output to NaN values.

    Input:
    - df: pd.dataFrame(index=DatetimeIndex, columns = [load1, ..., loadN]). Load_corrections should be indicated by 'LC_'
    - suspicious_moments (pd.dataframe): output of function nonzero_flatliner in new variable


    return:
    - pd.DataFrame without invalid data (converted to NaN values)"""
    if suspicious_moments is not None:
        for index, row in suspicious_moments.iterrows():
            df[(row[0]) : (row[1])] = np.nan
        return df
    return df



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from pathlib import Path

import pymsteams
from ktpbase.config.config import ConfigManager
from ktpbase.log import logging


def post_teams(msg, invalid_coefs=None, coefsdf=None, url=None):
    """Post a message to Teams - KTP

    Note that currently no authentication occurs.
    Security is given by keeping the URL secret.
    One should therefore refrain from using more enhanced features such as
    action buttons.

    Args:
        msg (mixed): For simple messages a string can be passed. For more
            complex messages pass a dict. The following keys are supported:
            text, links, sections. Each section can contain the following keys:
            text, title, images, facts, markdown. Also see:
            https://docs.microsoft.com/en-us/outlook/actionable-messages/send-via-connectors
        invalid_coefs (pd.DatFrame, optional): df of information of invalid
            coefficients. Defaults to None.
        coefsdf (pd.DataFrame, optional): df of new coefficients. Defaults to None.
        url (string, optional): webhook url, monitoring by default

    Note:
        This function is namespace-specific.

    """
    config = ConfigManager.get_instance()
    logger = logging.get_logger(__name__)

    # If no url is passed fall back to default
    if url is None:
        # if Teams url is not configured just return
        if hasattr(config, "teams") is False or config.teams.monitoring_url is None:
            logger.warning("Can't post Teams message, no url given")
            return
        else:
            logger.info("No url given, using default from config")
            url = config.teams.monitoring_url

    card = pymsteams.connectorcard(url)

    # add proxies
    # NOTE the connectorcard.proxy is passed to the requests library under the hood
    card.proxies = config.proxies

    # if msg is string, convert to dict
    if type(msg) is str:
        msg = dict(text=msg)
    card.text(msg.get("text"))
    card.summary(msg.get("fallback", "-"))

    # set title, color, ...
    card.color(msg.get("color", "white"))
    card.title(msg.get("title"))

    for link_dict in msg.get("links", []):
        card.addLinkButton(link_dict["buttontext"], link_dict["buttonurl"])

    # Add sections
    for section_dict in msg.get("sections", []):
        section = pymsteams.cardsection()

        section.text(section_dict.get("text"))
        section.title(section_dict.get("title"))
        for image in section_dict.get("images", []):
            section.addImage(image)
        for fact in section_dict.get("facts", []):
            section.addFact(*fact)
        if not section_dict.get("markdown", True):
            section.disableMarkdown()
        if "link" in section_dict:
            section.linkButton(
                section_dict.get("link").get("buttontext"),
                section_dict.get("link").get("buttonurl"),
            )

        card.addSection(section)

    card.send()


def post_teams_alert(msg, url=None):
    """Same as post_teams, but posts to alert channel.

    Args:
        msg (mixed): For simple messages a string can be passed. For more
            complex messages pass a dict. The following keys are supported:
            text, links, sections. Each section is a dict and can contain the
            following keys: text, title, images, facts, markdown. Also see:
            https://docs.microsoft.com/en-us/outlook/actionable-messages/send-via-connectors

    Note:
        This function is namespace-specific.
    """
    config = ConfigManager.get_instance()
    if url is None:
        if hasattr(config, "teams") is True:
            url = config.teams.alert_url

    return post_teams(msg, url=url)


def build_sql_query_string(df, table):
    """Build sql insert query string for Teams message output from df.

    Args:
        df (pd.DataFrame): df of table values to insert in sql
        table (string): table to insert df into

    Returns:
        string: sql query string of insert statement
    """
    # round all values to two decimals
    df = df.round(2)
    # convert datetime to string format
    datetime_columns = df.columns[
        df.columns.isin(["date_start", "date_end", "created"])
    ]
    for col in datetime_columns:
        df[col] = df[col].astype("str")

    sql_texts = [
        "```  \nINSERT INTO "
        + table
        + " ("
        + str(", ".join(df.columns))
        + ") VALUES  \n"
    ]
    for index, row in df.iterrows():
        if index != df.index[0]:
            sql_texts.append(",  \n")  # 2 spaces and \n create a new line
        sql_texts.append(str(tuple(row.values)))
    sql_texts.append("  \n```")
    query = "".join(sql_texts)
    return query


def send_report_teams_better(pj, feature_importance):
    """Send a report to teams for monitoring input for an improved model.

    Post includes information (performance, figures, etc.) about the trained
    model. Use when the new trained model is better than the old model.

    Args:
        pj (dict): A dictionarry specifying the prediction job. This dict should
            at least contain the following keys: {
                'id': (int),
                'sid': (str),
                'name': (str),
                'horizon_minutes': (int),
                'resolution_minutes': (int),
                'lat': (float),
                'lon': (float),
                'description': (str)
            }.
            Usually this dictionary results from querrying the 'predictions'
            table in the SQL database.
        feature_importance (pandas.DataFrame): A DataFrame describing the
            feature importances and weights of the trained model.
    model: XGBoost model object of the newly trained model

    Returns:
        None

    """
    config = ConfigManager.get_instance()
    web_link = f'{config.dashboard.trained_models_url}/{pj["id"]}'

    msg = {
        "fallback": f'Trained better model: {pj["name"]}',
        "title": "Trained better model",
        "sections": [
            {
                "facts": [
                    ("Name", pj["name"]),
                    ("Desc", pj["description"]),
                    ("pid", pj["id"]),
                ],
                "markdown": False,
            },
            {
                "title": "Dominant features",
                "facts": [
                    (
                        feature_importance.index[0],
                        f'{feature_importance["gain"][0]:.1%}',
                    ),
                    (
                        feature_importance.index[1],
                        f'{feature_importance["gain"][1]:.1%}',
                    ),
                ],
            },
        ],
        "links": [
            {
                "buttontext": "Train Performance",
                "buttonurl": f"{web_link}/Predictor47.0.html",
            },
            {
                "buttontext": "Model Weights",
                "buttonurl": f"{web_link}/weight_plot.html",
            },
        ],
        "color": "#764FA5",
    }

    post_teams(msg)


def send_report_teams_worse(pj):
    """Send a report to teams for monitoring input for a worsened model.

    Post includes information (performance, figures, etc.) about the trained
    model. Use when the new trained model is worse than the old model.

    Args:
        pj (dict): A dictionarry specifying the prediction job. This dict should
            at least contain the following keys: {
                'id': (int),
                'sid': (str),
                'name': (str),
                'horizon_minutes': (int),
                'resolution_minutes': (int),
                'lat': (float),
                'lon': (float),
                'description': (str)
            }.
            Usually this dictionary results from querrying the 'predictions'
            table in the SQL database.
    model: XGBoost model object of the newly trained model

    Returns:
        None

    """
    config = ConfigManager.get_instance()
    web_link_old = f'{config.dashboard.trained_models_url}/{pj["id"]}'
    web_link_new = f'{config.dashboard.trained_models_url}/{pj["id"]}/worse_model'

    image_save_location = Path(config.paths.trained_models) / f'{pj["id"]}'

    with open(image_save_location / "worse_model" / "Predictor47.0.datauri", "rt") as f:
        graph = f.read()

    msg = {
        "fallback": f'Trained worse model: {pj["name"]}',
        "title": "Warning",
        "text": "Old model is better. Please check and retrain using tracy if necessary.",
        "sections": [
            {
                "facts": [
                    ("Name", pj["name"]),
                    ("Desc", pj["description"]),
                    ("pid", pj["id"]),
                ],
                "markdown": False,
            },
            {
                "title": "New Model Performance",
                "images": [graph],
            },
        ],
        "links": [
            {
                "buttontext": "Old Model Performance",
                "buttonurl": f"{web_link_old}/Predictor47.0.html",
            },
            {
                "buttontext": "New Model Performance",
                "buttonurl": f"{web_link_new}/Predictor47.0.html",
            },
            {
                "buttontext": "Old Model Weights",
                "buttonurl": f"{web_link_old}/weight_plot.html",
            },
            {
                "buttontext": "New Model Weights",
                "buttonurl": f"{web_link_new}/weight_plot.html",
            },
        ],
        "color": "#a5764f",
    }

    post_teams(msg)



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

# -*- coding: utf-8 -*-
"""optimize_hyper_params.py

This module contains the CRON job that is periodically executed to optimize the
hyperparameters for the prognosis models.

Example:
    This module is meant to be called directly from a CRON job. A description of
    the CRON job can be found in the /k8s/CronJobs folder.
    Alternatively this code can be run directly by running::

        $ python optimize_hyper_params.py

"""
from openstf.model.hyper_parameters import optimize_hyperparameters_pj
from openstf.tasks.utils.predictionjobloop import PredictionJobLoop
from openstf.tasks.utils.taskcontext import TaskContext


def main():
    with TaskContext(__file__) as context:
        model_type = ["xgb", "xgb_quantile", "lgb"]

        PredictionJobLoop(context, model_type=model_type).map(
            optimize_hyperparameters_pj
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

""" This module contains all wheather related functions used for feature engineering.

"""
import numpy as np
import pandas as pd


# Set some (nameless) constants for the Antoine equation:
A = 6.116
M = 7.6
TN = 240.7
# Set some constants
TORR = 133.322368  # 1 torr = 133 Pa
# 1.168 is the mass of 1 m^3 of air on sea level with standard pressure.
D = 1.168


def calc_saturation_pressure(temperature):
    """Function that calculates the water vapour pressure from the temperature
    See https://www.vaisala.com/sites/default/files/documents/Humidity_Conversion_Formulas_B210973EN-F.pdf

    Args:
        Temperature (np.array): Temperature in C
    Returns:
        The saturation pressure of water at the respective temperature"""

    psat = A * 10 ** ((M * temperature) / (temperature + TN))
    return psat


def calc_vapour_pressure(rh, psat):
    """Calculates the vapour pressure

    Args:
        rh (np.ndarray or float): Relative humidity
        psat (np.ndarray or float): Saturation pressure: see calc_saturation_pressure
    Returns:
        The water vapour pressure"""
    return (rh) * psat


def calc_dewpoint(vapour_pressure):
    """Calculates the dewpoint

    Args:
        vapour_pressure (np.ndarray or float): The vapour pressure for which the dewpoint should be calculated
    Returns:
        dewpoint (np.ndarray or float):"""

    return TN / ((M / np.log10(vapour_pressure / A)) - 1)


def calc_air_density(temperature, pressure, rh):
    """Calculates the dewpoint

    Args:
        Temperature (np.ndarray or float): The temperature in C
        Pressure (np.ndarray or float): the atmospheric pressure in Pa
    Returns:
        Air density (np.ndarray or float): The air density (kg/m^3)"""

    # Calculate saturation pressure
    psat = calc_saturation_pressure(temperature)
    # Calculate the current vapour pressure
    vapour_pressure = calc_vapour_pressure(rh, psat)

    # Set tempareture to K
    temperature_k = temperature + 273.15

    # Calculate air density
    air_density = (
        D
        * (273.15 / temperature_k)
        * ((pressure - 0.3783 * vapour_pressure) / 760 / TORR)
    )

    return air_density


def humidity_calculations(temperature, rh, pressure):
    """Function that calculates the
    - Saturation pressure
    - Vapour pressure
    - Dewpoint
    - Air density
    Args:
        temperature (np.array): Temperature in C
        rh (np.array): Relative humidity in %
        pressure (np.array): The air pressure in hPa
    Returns:
        if the input is an np.ndarray: a pandas dataframe with the calculated moisture indices
        if the input is numeric: a dict with the calculated moisture indices"""

    # First: a sanity check on the relative humidity and the air pressure
    # We only check on the type of temperature, because they need to be the same anyway
    is_series = isinstance(temperature, (np.ndarray, pd.core.series.Series))
    is_scalar = isinstance(temperature, (float, int))

    if is_scalar is False and is_series is False:
        raise TypeError(
            "The input should be a pandas series or np.ndarry, or float or int"
        )

    # Suppres copy warnings
    with pd.option_context("mode.chained_assignment", None):
        if is_series:
            rh[rh > 1] = rh / 100  # This triggers copy warnings
            pressure[pressure < 80000] = np.nan  # This triggers copy warnings
        else:
            if rh > 1:
                rh /= 100
            if pressure < 80000:
                pressure = np.nan

    # If the input is a dataframe or np.ndarrays: return a dataframe
    if is_series:
        humidity_df = pd.DataFrame(
            columns=[
                "saturation_pressure",
                "vapour_pressure",
                "dewpoint",
                "air_density",
            ]
        )
        humidity_df["saturation_pressure"] = calc_saturation_pressure(temperature)
        humidity_df["vapour_pressure"] = calc_vapour_pressure(
            rh, humidity_df.saturation_pressure
        )
        humidity_df["dewpoint"] = calc_dewpoint(humidity_df.vapour_pressure)
        humidity_df["air_density"] = calc_air_density(temperature, pressure, rh)

        return humidity_df

    # Else: if the input is numeric: return a dict
    psat = calc_saturation_pressure(temperature)
    pw = calc_vapour_pressure(rh, psat)
    td = calc_dewpoint(pw)
    air_density = calc_air_density(temperature, pressure, rh)
    return {
        "saturation_pressure": psat,
        "vapour_pressure": pw,
        "dewpoint": td,
        "air_density": air_density,
    }



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import pickle

import xgboost as xgb

from openstf.model.serializer.serializer import AbstractModelSerializer


class XGBModelSerializer(AbstractModelSerializer):
    def __init__(self):
        super().__init__()

    def save(self, pid, xgb_model, corrections=None):

        save_folder = self.build_save_folder_path(pid=pid)

        save_folder.mkdir(parents=True, exist_ok=True)

        # save XGB model
        xgb_model.save_model(save_folder / self.MODEL_FILENAME)
        # save feature names
        with open(save_folder / self.FEATURE_NAMES_FILENAME, "w") as fh:
            fh.write(str(xgb_model.feature_names))
        # save feature types
        with open(save_folder / self.FEATURE_TYPES_FILENAME, "w") as fh:
            fh.write(str(xgb_model.feature_types))
        # save best iteration / best n_trees
        if getattr(xgb_model, "best_ntree_limit") is not None:
            with open(save_folder / self.BEST_ITERATION_FILENAME, "wb") as fh:
                pickle.dump(xgb_model.best_ntree_limit, fh)
        else:
            self.logger.warning(
                f"No best iteration found, no best iteration save for pid: {pid}"
            )
        # save corrections
        if corrections is not None:
            if len(corrections) == len(corrections.dropna()):
                corrections.to_csv(
                    save_folder / self.CORRECTIONS_FILENAME, **self.PANDAS_CSV_FORMAT
                )
            return

        self.logger.warning(
            "No corrections found, corrections file not saved for pid: {pid}"
        )

    # NOTE from general.py
    def load(self, pid, pid_model_folder=None):
        """Load the most recent model.

        Args:
            pid (int): Prediction job id
            pid_model_folder(str): Path to save model at specific location

        Raises:
            FileNotFoundError: If no recent model file was found

        Returns:
            tuple: Tuple with:
                [0]: xgb.Booster: Loaded model
                [1]: str: Path to loaded model
        """

        if pid_model_folder is None:
            pid_model_folder = self.build_pid_model_folder_path(pid=pid)

        try:
            model_file, model_folder = self.find_most_recent_model_file(
                pid_model_folder
            )
        except FileNotFoundError as e:
            self.logger.error(f"Can't load model, no recent model file found. {e}")
            return None, None

        xgb_model = xgb.Booster(model_file=model_file)

        self.logger.info(f"Loaded model from: {model_file}")

        xgb_model = self._add_atributes(model_folder, xgb_model)

        return xgb_model, model_file

    def _add_atributes(self, model_folder, model):

        # load feature names
        with open(model_folder / self.FEATURE_NAMES_FILENAME, "r") as fh:
            # Reformat feature names
            # TODO this is unnecessary when just saved properly (i.e yaml, json, xml, ini, pickle etc)
            model.feature_names = (
                fh.read()
                .replace("[", "")
                .replace("]", "")
                .replace("'", "")
                .replace(" ", "")
                .split(",")
            )
        # Load feature types
        with open(model_folder / self.FEATURE_TYPES_FILENAME, "r") as fh:
            # Reformat feature types
            model.feature_types = (
                fh.read()
                .replace("[", "")
                .replace("]", "")
                .replace("'", "")
                .replace(" ", "")
                .split(",")
            )

        # Load best iteration (if exist)
        best_iteration_filepath = model_folder / self.BEST_ITERATION_FILENAME
        if best_iteration_filepath.is_file() is True:
            with open(best_iteration_filepath, "rb") as fh:
                model.best_ntree_limit = pickle.load(fh)
        else:
            self.logger.warning(
                f"Could not load best iteration, file does not exists '{best_iteration_filepath}'"
            )
            # TODO magic number?
            model.best_ntree_limit = 15
        return model



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

# -*- coding: utf-8 -*-
"""apply_features.py.

This module provides functions for applying features to the input data.
This improves forecast accuracy. Examples of features that are added are:
    The load 1 day and 7 days ago at the same time
    If a day is a weekday or a holiday
    The extrapolated windspeed at 100m
    The normalised wind power according to the turbine-specific power curve

"""

import numpy as np
import pandas as pd
import scipy.signal

from openstf.feature_engineering.feature_free_days import create_holiday_functions
from openstf.feature_engineering.weather import humidity_calculations


def generate_lag_functions(data, minute_list=None, h_ahead=24):
    """Creates functions to generate lag features in a dataset.

    Args:
        data (pd.DataFrame): input data for an xgboost prediction or model training.
        minute_list (list of ints): minute lagtimes that where used during training
            of the model. If empty a new et will be automatically generated.
        h_ahead (int): Forecast horizon limit in hours.

    Returns:
        dict: dictionary with lag functions

    Example:
        lag_functions = generate_lag_functions(data,minute_list,h_ahead)
    """

    if minute_list is None:
        minute_list = []

    ##########
    # Define feature function groups
    lag_functions = {}
    # Add intraday-lags
    if h_ahead < 24:
        minminutes = int(np.round(np.ceil(h_ahead) * 60 / 15) * 15)
        minutespace = np.linspace(
            minminutes, 23 * 60, int(23 - minminutes / 60 + 1)
        ).tolist()

        # if no aditional lag time feature are defined investigate if we should add some
        if minute_list == []:
            minute_list = additional_minute_space(data)

        # add intra-hour values if hAhead > 1
        if h_ahead < 1:
            minutespace = set([15, 30, 45] + minutespace + minute_list)
        for minutes in minutespace:

            def func(x, shift=minutes):
                return x.shift(freq="T", periods=1 * shift)

            new = {"T-" + str(int(minutes)) + "min": func}
            lag_functions.update(new)

    # Define time lags in days:
    mindays = int(np.ceil(h_ahead / 24))
    for day in np.linspace(mindays, 14, 15 - mindays):

        def func(x, shift=day):
            return x.shift(freq="1d", periods=1 * shift)

        new = {"T-" + str(int(day)) + "d": func}
        lag_functions.update(new)
    return lag_functions


def additional_minute_space(data, height_treshold=0.1):
    """This script calculates an autocorrelation curve of the load trace. This curve is
        subsequently used to add additional lag times as features.

    Args:
        data (pandas.DataFrame): a pandas dataframe with input data in the form pd.DataFrame(index = datetime,
                             columns = [label, predictor_1,..., predictor_n])
        height_treshold (float): minimal autocorrelation value to be recognized as a peak.

    Returns:
        list of ints with aditional minute lags


    """

    def autocorr(x, lags):
        """Function to make a autocorrelation curve"""
        mean = x.mean()
        var = np.var(x)
        xp = x - mean
        corr = np.correlate(xp, xp, "full")[len(x) - 1 :] / var / len(x)

        return corr[: len(lags)]

    try:
        # Get rid of nans as the autocorrelation handles these values badly
        data = data[data.columns[0]].dropna()
        # Get autocorrelation curve
        y = autocorr(data, range(10000))
        # Determine the peaks (positive and negative) larger than a specified threshold
        peaks = scipy.signal.find_peaks(np.abs(y), height=height_treshold)
        peaks = peaks[0]
        # Convert peaks to lag times in minutes
        peaks = peaks[peaks < (60 * 4)]
        additional_minute_space = peaks * 15
    except Exception:
        return []
    # Return list of additional minute lags to be procceses by apply features
    return list(additional_minute_space)


def apply_features(data, minute_list=None, h_ahead=24):
    """This script applies the feature functions defined in
        feature_functions.py and returns the complete dataframe. Features requiring
        more recent label-data are omitted.
    Args:
        data (pandas.DataFrame): a pandas dataframe with input data in the form:
                                    pd.DataFrame(
                                        index=datetime,
                                        columns=[label, predictor_1,..., predictor_n]
                                    )
        minute_list (list of ints): minute lagtimes that where used during training of
                                    the model. If empty a new et will be automatically
                                    generated.
        h_ahead (int): Forecast horizon limit in hours.

    Returns:
        pd.DataFrame(index = datetime, columns = [label, predictor_1,..., predictor_n,
            feature_1, ..., feature_m])

    Example:
        import pandas as pd
        import numpy as np
        index = pd.date_range(start = "2017-01-01 09:00:00",
        freq = '15T', periods = 200)
        data = pd.DataFrame(index = index,
                            data = dict(load=
                            np.sin(index.hour/24*np.pi)*
                            np.random.uniform(0.7,1.7, 200)))

    """
    if minute_list is None:
        minute_list = []

    lag_functions = generate_lag_functions(data, minute_list, h_ahead)

    # TODO it seems not all feature use the same convention
    # better to choose one and stick to it
    timedriven_functions = {
        "IsWeekendDay": lambda x: (x.index.weekday // 5) == 1,
        "IsWeekDay": lambda x: x.index.weekday < 5,
        "IsMonday": lambda x: x.index.weekday == 0,
        "IsTuesday": lambda x: x.index.weekday == 1,
        "IsWednesday": lambda x: x.index.weekday == 2,
        "IsThursday": lambda x: x.index.weekday == 3,
        "IsFriday": lambda x: x.index.weekday == 4,
        "IsSaturday": lambda x: x.index.weekday == 5,
        "IsSunday": lambda x: x.index.weekday == 6,
        "Month": lambda x: x.index.month,
        "Quarter": lambda x: x.index.quarter,
    }

    # Add check for specific hour
    for thour in np.linspace(0, 23, 24):

        def func(x, checkhour=thour):
            return x.index.hour == checkhour

        char = "0" if thour < 10 else ""
        new = {"Is" + char + str(int(thour)) + "Hour": func}
        timedriven_functions.update(new)

    df = data.copy()
    # holiday_function are imported at the beginning of the file and includes all dutch
    # school/work holidays
    holiday_functions = create_holiday_functions()

    for function_group in [lag_functions, timedriven_functions, holiday_functions]:
        for name, featfunc in function_group.items():
            df[name] = df.iloc[:, [0]].apply(featfunc)

    if "windspeed" in list(df):
        df["windspeed_100mExtrapolated"] = calculate_windspeed_at_hubheight(
            df.windspeed
        )
        df["windPowerFit_extrapolated"] = calculate_windturbine_power_output(
            df["windspeed_100mExtrapolated"]
        )

    if "windspeed_100m" in list(df):
        df["windpowerFit_harm_arome"] = calculate_windturbine_power_output(
            df["windspeed_100m"].astype(float)
        )

    # Try to add humidity  calculations, ignore if required columns are missing
    try:
        humidity_df = humidity_calculations(df.temp, df.humidity, df.pressure)
        df = df.join(humidity_df)
    except AttributeError:
        pass  # This happens when a required column for humidity_calculations
        # is not present

    return df


def apply_multiple_horizon_features(data, h_aheads=[0.25, 24], latency={"APX": 24}):
    """Based on ApplyFeatures, but suited for generating a table containing
    multiple time horizons at once./
    Also take function latency into consideration

    Input:
        - data: pd.DataFrame(index = datetime, columns =
                             [label, predictor_1,..., predictor_n])
        - hAhead: list. Forecast horizon limit in hours.
        Features requiring more recent label-data are omitted.
        - latency: dict(key=value). set data[key] to nan for horizon > value

    Output:
        - pd.DataFrame(index = datetime, columns =
            [label, predictor_1,..., predictor_n, feature_1, ..., feature_m])

    Example
    import pandas as pd
    import numpy as np
    index = pd.date_range(start = "2017-01-01 09:00:00",
                             freq = '15T', periods = 200)
    data = pd.DataFrame(index = index,
                        data = dict(load=np.sin(index.hour/24*np.pi)*
                                    np.random.uniform(0.7,1.7, 200)))"""

    result = pd.DataFrame()
    cols = []
    for h_ahead in h_aheads:
        res = apply_features(data.copy(), h_ahead=h_ahead)
        res["Horizon"] = h_ahead
        if len(res.columns) > len(cols):
            cols = res.columns
        result = result.append(res, sort=False)  # appending unsorts columns

    # apply data latency
    for feature, time in latency.items():
        result.loc[result["Horizon"] > time, feature] = np.nan

    return result[cols].sort_index()


def calculate_windspeed_at_hubheight(windspeed, fromheight=10, hub_height=100):
    """
    function that extrapolates a wind from a certain height to 100m
    According to the wind power law (https://en.wikipedia.org/wiki/Wind_profile_power_law)

    input:
        - windspeed: float OR pandas series of windspeed at height = height
        - fromheight: height (m) of the windspeed data. Default is 10m
        - hubheight: height (m) of the turbine
    returns:
        - the windspeed at hubheight."""
    alpha = 0.143

    if not isinstance(windspeed, (np.ndarray, float, int, pd.core.series.Series)):
        raise TypeError(
            "The windspeed is not of the expected type!\n\
                        Got {}, expected np.ndarray, pd series or numeric".format(
                type(windspeed)
            )
        )

    try:
        if any(windspeed < 0):
            raise ValueError(
                "The windspeed cannot be negative, as it is the lenght of a vector"
            )
    except TypeError:
        if windspeed < 0:
            raise ValueError(
                "The windspeed cannot be negative, as it is the lenght of a vector"
            )
        windspeed = abs(windspeed)

    return windspeed * (hub_height / fromheight) ** alpha


def calculate_windturbine_power_output(windspeed, n_turbines=1, turbine_data=None):
    """This function calculates the generated wind power based on the wind speed.
    These values are related through the power curve, which is described by turbine_data.
    If no turbine_data is given, default values are used and results are normalized to 1MWp.
    If n_turbines=0, the result is normalized to a rated power of 1.

    Input:
        - windspeed: pd.DataFrame(index = datetime, columns = ["windspeedHub"])
        - nTurbines: int
        - turbineData: dict(slope_center, rated_power, steepness)

    Ouput:
        pd.DataFrame(index = datetime, columns = ["forecast"])"""

    if turbine_data is None:
        turbine_data = {
            "name": "Lagerwey L100",  # not used here
            "cut_in": 3,  # not used here
            "cut_off": 25,  # not used here
            "kind": "onshore",  # not used here
            "manufacturer": "Lagerwey",  # not used here
            "peak_capacity": 1,  # not used here
            "rated_power": 1,
            "slope_center": 8.07,
            "steepness": 0.664,
        }
    else:
        required_properties = ["rated_power", "steepness", "slope_center"]
        for prop in required_properties:
            if prop not in turbine_data.keys():
                raise KeyError(f"Required property '{prop}' not set in turbine data")

    generated_power = turbine_data["rated_power"] / (
        1
        + np.exp(
            -turbine_data["steepness"] * (windspeed - turbine_data["slope_center"])
        )
    )
    generated_power *= n_turbines

    return generated_power



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

# This file contains the Objective loss functions for quantile regression from:
# https://gist.github.com/Nikolay-Lysenko/06769d701c1d9c9acb9a66f2f9d7a6c7
#
# SPDX-FileCopyrightText: 2017 Nikolay Lysenko
#
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd


def rmse(realised, forecast):
    """Function that calculates the Root Mean Square Error
    based on the true and prediciton"""
    return np.sqrt(((realised - forecast) ** 2).mean())


def bias(realised, forecast):
    """Function that calculates the absolute bias in %
    based on the true and prediciton"""
    return np.mean(forecast - realised)


def nsme(realised, forecast):
    """Function that calculates the Nash-sutcliffe model efficiency
    based on the true and prediciton"""
    try:
        return 1 - sum((forecast - realised) ** 2) / sum(
            (realised - np.mean(realised)) ** 2
        )
    except ZeroDivisionError:  # means the error is 0
        return 1


def mae(realised, forecast):
    """Function that calculates the mean absolute error
    based on the true and prediction"""
    return np.mean(np.abs(forecast - realised))


def r_mae(realised, forecast):
    """Function that calculates the relative mean absolute error
    based on the true and prediction.
    The range is based on the load range of the previous two weeks"""

    # Determine load range on entire dataset
    range_ = realised.max() - realised.min()

    return mae(realised, forecast) / range_


def frac_in_stdev(realised, forecast, stdev):
    """Function that calculates the amount of measurements that are within one stdev of
    our predictions"""
    outside_stdev = forecast[(forecast - realised).abs() > stdev]
    return round((1 - (len(outside_stdev) / len(forecast))), 2)


def r_mae_highest(realised, forecast, percentile=0.95):
    """Function that calculates the relative mean absolute error
    based on the true and prediction for the 5 percent highest realised values.
    The range is based on the load range of the previous two weeks"""

    # Check if length of both arrays is equal
    if len(np.array(realised)) != len(np.array(forecast)):
        raise ValueError(
            "Error metric can only be calculated for arrays of equal length!"
        )

    # Determine load range on entire dataset
    range_ = realised.max() - realised.min()

    # Get highest percentile of values
    highest_values = realised > np.percentile(realised, percentile)

    # Calculate mae
    r_mae_highest = mae(realised[highest_values], forecast[highest_values]) / range_

    return r_mae_highest


def r_mne_highest(realised, forecast):
    """Function that calculates the relative mean negative error
    based on the true and prediction for the 5 percent highest realised values.
    The range is based on the load range of the previous two weeks, this measure
    quantifies how much we underestimate peaks"""

    # Combine series in one DataFrame
    combined = pd.concat([realised, forecast], axis=1)

    # Determine load range on entire dataset
    range_ = combined["load"].max() - combined["load"].min()

    # Select 5 percent highest realised load values
    combined["highest"] = combined["load"][
        combined["load"] > combined["load"].quantile(0.95)
    ]
    combined = combined[np.invert(np.isnan(combined["highest"]))]

    # Calculate rMNE for the selected points
    diff = combined[forecast.name] - combined["load"]

    if len(diff[diff < 0]) < 2:
        return 0.0

    r_mne_highest = np.mean(diff[diff < 0]) / range_

    if np.isnan(r_mne_highest):
        return 99999.0

    return r_mne_highest


def r_mpe_highest(realised, forecast):
    """Function that calculates the relative mean positive error
    based on the true and prediction for the 5 percent highest realised values.
    The range is based on the load range of the previous two weeks, this measure
    quantifies how much we overestimate peaks
    """

    # Combine series in one DataFrame
    combined = pd.concat([realised, forecast], axis=1)

    # Determine load range on entire dataset
    range_ = combined["load"].max() - combined["load"].min()

    # Select 5 percent highest realised load values
    combined["highest"] = combined["load"][
        combined["load"] > combined["load"].quantile(0.95)
    ]
    combined = combined[np.invert(np.isnan(combined["highest"]))]

    # Calculate rMPE for the selected points

    diff = combined[forecast.name] - combined["load"]

    if len(diff[diff > 0]) < 2:
        return 0.0

    r_mpe_highest = np.mean(diff[diff > 0]) / range_

    if np.isnan(r_mpe_highest):
        return 99999.0
    return r_mpe_highest


def r_mae_lowest(realised, forecast, quantile=0.05):
    """Function that calculates the relative mean absolute error
    based on the true and prediction for the 5 percent lowest realised values.
    The range is based on the load range of the previous two weeks"""

    # Determine load range on entire dataset
    range_ = realised.max() - realised.min()

    # Get lowest percentile of values
    lowest_values = realised < np.quantile(realised, quantile)
    # Calculate mae
    r_mae_lowest = mae(realised[lowest_values], forecast[lowest_values]) / range_

    return r_mae_lowest


def skill_score(realised, forecast, mean):
    """Function that calculates the skill score.

    Thise indicates model performance relative to a reference, in this case the mean
    of the realised values. The range is based on the load range of the previous
    two weeks
    """
    combined = pd.concat([realised, forecast], axis=1)
    combined["mean"] = mean

    skill_score = 1 - (mae(realised, forecast) / mae(realised, combined["mean"]))

    if np.isnan(skill_score):
        return 0

    return skill_score


def skill_score_positive_peaks(realised, forecast, mean):
    # Combine series in one DataFrame
    combined = pd.concat([realised, forecast], axis=1)

    # Select 5 percent highest realised load values
    combined["highest"] = combined["load"][
        combined["load"] > combined["load"].quantile(0.95)
    ]
    combined = combined[np.invert(np.isnan(combined["highest"]))]

    # Calculate rMAE for the selected points
    skill_score_highest = skill_score(combined["load"], combined[forecast.name], mean)

    if np.isnan(skill_score_highest):
        return 0

    return skill_score_highest


def franks_skill_score(realised, forecast, basecase, range_=1.0):
    # Combine series in one DataFrame
    combined = pd.concat([realised, forecast], axis=1)
    if range_ == 1.0:
        range_ = combined["load"].max() - combined["load"].min()

    franks_skill_score = (mae(realised, basecase) - mae(realised, forecast)) / range_

    if np.isnan(franks_skill_score):
        return 0

    return franks_skill_score


def franks_skill_score_peaks(realised, forecast, basecase):
    # Combine series in one DataFrame
    combined = pd.concat([realised, forecast, basecase], axis=1)

    range_ = combined["load"].max() - combined["load"].min()
    # Select 5 percent highest realised load values
    combined["highest"] = combined["load"][
        combined["load"] > combined["load"].quantile(0.95)
    ]
    combined = combined[np.invert(np.isnan(combined["highest"]))]

    # Calculate rMAE for the selected points
    franks_skill_score_highest = franks_skill_score(
        combined["load"],
        combined[forecast.name],
        combined[basecase.name],
        range_=range_,
    )

    if np.isnan(franks_skill_score_highest):
        return 0

    return franks_skill_score_highest


# Objective loss functions for quantile regression, from: https://gist.github.com/Nikolay-Lysenko/06769d701c1d9c9acb9a66f2f9d7a6c7

# SPDX-FileCopyrightText: 2017 Nikolay Lysenko
#
# SPDX-License-Identifier: MIT


def xgb_quantile_eval(preds, dmatrix, quantile=0.2):
    """
    Customized evaluational metric that equals
    to quantile regression loss (also known as
    pinball loss).
    Quantile regression is regression that
    estimates a specified quantile of target's
    distribution conditional on given features.
    @type preds: numpy.ndarray
    @type dmatrix: xgboost.DMatrix
    @type quantile: float
    @rtype: float

    # See also:
    https://gist.github.com/Nikolay-Lysenko/06769d701c1d9c9acb9a66f2f9d7a6c7

    """
    labels = dmatrix.get_label()
    return (
        "q{}_loss".format(quantile),
        np.nanmean(
            (preds >= labels) * (1 - quantile) * (preds - labels)
            + (preds < labels) * quantile * (labels - preds)
        ),
    )


def xgb_quantile_obj(preds, dmatrix, quantile=0.2):
    """
    Computes first-order derivative of quantile
    regression loss and a non-degenerate
    substitute for second-order derivative.
    Substitute is returned instead of zeros,
    because XGBoost requires non-zero
    second-order derivatives. See this page:
    https://github.com/dmlc/xgboost/issues/1825
    to see why it is possible to use this trick.
    However, be sure that hyperparameter named
    `max_delta_step` is small enough to satisfy:
    ```0.5 * max_delta_step <=
       min(quantile, 1 - quantile)```.
    @type preds: numpy.ndarray
    @type dmatrix: xgboost.DMatrix
    @type quantile: float
    @rtype: tuple(numpy.ndarray)

    # See also:
    https://gist.github.com/Nikolay-Lysenko/06769d701c1d9c9acb9a66f2f9d7a6c7

    Reasoning for the hessian:
    https://gist.github.com/Nikolay-Lysenko/06769d701c1d9c9acb9a66f2f9d7a6c7#gistcomment-2322558
    """
    try:
        assert 0 <= quantile <= 1
    except AssertionError:
        raise ValueError("Quantile value must be float between 0 and 1.")

    labels = dmatrix.get_label()
    errors = preds - labels

    left_mask = errors < 0
    right_mask = errors > 0

    grad = -quantile * left_mask + (1 - quantile) * right_mask
    hess = np.ones_like(preds)

    return grad, hess



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from openstf.model.general import MLModelType
from openstf.model.trainer.xgboost.quantile import XGBQuantileModelTrainer
from openstf.model.trainer.xgboost.xgboost import XGBModelTrainer
from openstf.model.trainer.lightgbm.lightgbm import LGBModelTrainer


class ModelTrainerCreator:
    """Factory object for creating model trainer objects"""

    # Set object mapping
    MODEL_TRAINER_CONSTRUCTORS = {
        MLModelType.XGB: XGBModelTrainer,
        MLModelType.XGB_QUANTILE: XGBQuantileModelTrainer,
        MLModelType.LGB: LGBModelTrainer,
    }

    def __init__(self, pj):
        # check if model type is valid
        if pj["model"] not in [k.value for k in self.MODEL_TRAINER_CONSTRUCTORS]:
            raise KeyError(f'Unknown model type: {pj["model"]}')

        self.pj = pj
        # TODO see if this can be configured more generally for example in a system yaml

    def create_model_trainer(self):
        """
        Method returns model trainer objects that can be used to train models

        Returns: Implementation of the AbstractModelTrainer class tailored to
            a specific algorithm.

        """
        model_type = MLModelType(self.pj["model"])
        model_trainer = self.MODEL_TRAINER_CONSTRUCTORS[model_type]

        return model_trainer(self.pj)



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime

import cufflinks as cf
import numpy as np
import pandas as pd

cf.go_offline()


def balance_classes(x, y):
    # find classes
    classes = y.unique()

    # find minimal number of samples per class
    n = min([len(np.unique(y[y == c].index.values)) for c in classes])

    # if min class is very small, raise amount of samples per class
    replace = False
    if n < 4:
        n = 4
        replace = True

    # sampled indices
    inds = []
    for c in classes:
        class_inds = np.unique(y[y == c].index.values)
        if len(class_inds) > 0:
            inds += list(np.random.choice(class_inds, size=n, replace=replace))

    mask = np.isin(y.index.values, inds)
    return x[mask], y[mask]


def prepare_training_data(df, y_col, val_n=4, val_width="14D"):
    # drop rows were load is not known
    df = df.dropna(subset=[y_col])

    # split in x (given) and y (to predict)
    x = df[[x for x in df.columns if x != y_col and x != "horizon"]]
    y = df[y_col]
    h = df["horizon"]

    # datetimes of data
    datetime_start = pd.to_datetime(str(df.index[0]))
    datetime_end = pd.to_datetime(str(df.index[-1]))

    # define validation set ranges
    val_ranges = []
    for i in range(val_n):
        val_start = (
            datetime_start
            + (i + 1) * ((datetime_end - datetime_start) / val_n)
            - pd.Timedelta(val_width)
        )
        val_end = val_start + pd.Timedelta(val_width)
        val_ranges.append((str(val_start.date()), str(val_end.date())))

    # validation set
    val_x = pd.concat([x[val_start:val_end] for val_start, val_end in val_ranges])
    val_y = pd.concat([y[val_start:val_end] for val_start, val_end in val_ranges])
    val_h = pd.concat([h[val_start:val_end] for val_start, val_end in val_ranges])

    # train set
    train_x = x.copy().drop(val_x.index)
    train_y = y.copy().drop(val_y.index)
    train_h = h.copy().drop(val_h.index)

    # balance classes
    train_x, train_y = balance_classes(train_x, train_y)
    val_x, val_y = balance_classes(val_x, val_y)

    return train_x, train_y, train_h, val_x, val_y, val_h


def prepare_prediction_data(df, y_col, y_hor):
    # pair prediction horizons to actual days
    horizon_dates = {}
    for h in y_hor:
        horizon_dates[str(h)] = str(
            (datetime.utcnow() + pd.Timedelta(str(h + 1) + "D")).date()
        )

    # for each horizon select corresponding input data
    x = []
    for h in horizon_dates:
        x.append(df.loc[horizon_dates[h]][df.loc[horizon_dates[h]]["horizon"] == h])
    x = pd.concat(x)

    # drop value to predict
    x = x.drop([y_col, "horizon"], axis="columns")

    return x


def visualize_predictions(df, classes):
    # define columns
    columns = []
    i = 0
    for c in classes:
        if i == 0:
            columns.append("<" + str(round(classes[c][1], 1)))
        elif i == len(classes) - 1:
            columns.append(str(round(classes[c][0], 1)) + str(">"))
        else:
            columns.append(
                str(round(classes[c][0], 1)) + "-" + str(round(classes[c][1], 1))
            )
        i += 1
    df.columns = columns

    # make plot
    fig = df.iplot(
        kind="heatmap",
        colorscale="blues",
        title="Capacity Prognosis (14D)",
        yTitle="Load Peak (MW)",
        asFigure=True,
        zmin=0,
        zmax=1,
    )

    return fig



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from openstf.model.general import MLModelType
from openstf.model.serializer.xgboost.quantile import XGBQuantileModelSerializer
from openstf.model.serializer.xgboost.xgboost import XGBModelSerializer
from openstf.model.serializer.lightgbm.lightgbm import LGBModelSerializer


class ModelSerializerCreator:

    MODEL_SERIALIZER_CONTRUCTORS = {
        MLModelType.XGB: XGBModelSerializer,
        MLModelType.XGB_QUANTILE: XGBQuantileModelSerializer,
        MLModelType.LGB: LGBModelSerializer,
    }

    @classmethod
    def create_model_serializer(cls, model_type):

        if model_type not in cls.MODEL_SERIALIZER_CONTRUCTORS:
            raise KeyError(f"Unkown model_type: '{model_type}'")

        return cls.MODEL_SERIALIZER_CONTRUCTORS[model_type]()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

"""split_forecast_into_components.py

This module contains the CRON job that is periodically executed to make
prognoses of solar features. These features are usefull for splitting the load
in solar and wind contributions.
This is achieved by carrying out the folowing steps:
  1. Get the wind and solar reference data for the specific location of the
     customer
  2. Get the TDCV (Typical Domestic Consumption Values) data
  3. Fit a linear combination of above time series to the historic load data to
     determine the contributions of each energy source.
  4. Write the resulting coeficients to the SQL database.

Example:
    This module is meant to be called directly from a CRON job. A description of
    the CRON job can be found in the /k8s/CronJobs folder.
    Alternatively this code can be run directly by running::

        $ python split_forecast_into_components.py

Attributes:


"""
from openstf.model.split_energy import split_energy
from openstf.tasks.utils.predictionjobloop import PredictionJobLoop
from openstf.tasks.utils.taskcontext import TaskContext


def main():
    with TaskContext(__file__) as context:
        model_type = ["xgb", "lgb"]

        PredictionJobLoop(
            context,
            model_type=model_type,
        ).map(lambda pj: split_energy(pj["id"]))


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path

from ktpbase.log import logging
from ktpbase.config.config import ConfigManager


class AbstractModelSerializer(ABC):

    MODEL_FILENAME = "model.bin"
    FOLDER_DATETIME_FORMAT = "%Y%m%d%H%M%S"
    CORRECTIONS_FILENAME = "corrections.csv"
    FEATURE_NAMES_FILENAME = "featurenames.txt"
    FEATURE_TYPES_FILENAME = "featuretypes.txt"
    BEST_ITERATION_FILENAME = "best_iteration.pkl"
    PANDAS_CSV_FORMAT = {"decimal": ".", "sep": ","}

    def __init__(self):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.trained_models_folder = Path(
            ConfigManager.get_instance().paths.trained_models
        )

    def __getstate__(self):
        # capture what is normally pickled
        state = self.__dict__.copy()
        # remove logger since its not pickable
        state.pop("logger")
        # what we return here will be stored in the pickle
        return state

    def __setstate__(self, newstate):
        # obtain a new logger
        newstate["logger"] = logging.get_logger(self.__class__.__name__)
        # re-instate our __dict__ state from the pickled state
        self.__dict__.update(newstate)

    @abstractmethod
    def load(self):
        pass

    @abstractmethod
    def save(self):
        pass

    def build_pid_model_folder_path(self, pid, custom_folder=None):
        """Build the path to the folder where trained models for the given pid are stored.

            The trainded models are stored a folder structure using the following
            template: <trained_models_folder>/<pid>[_<component-name>]/<YYYYMMDDHHMMSS>/

        Args:
            pid (int): Prediction id
            custom_folder (pathlike): Path to custom trainded models folder
        """

        model_folder_name = f"{pid}"

        if custom_folder is not None:
            trained_models_folder = custom_folder
        else:
            trained_models_folder = self.trained_models_folder

        model_folder = trained_models_folder / model_folder_name

        return model_folder

    def find_most_recent_model_file(self, pid_model_folder):
        """Find the model recent stored model file.

            Iterate over the directories in the 'pid' model folder (the top level model
            folder for a specific pid) and find the

        Args:
            pid_model_folder (pathlib.Path): [description]

        Raises:
            FileNotFoundError: When no model file is found in any of the subdirectories
                of `pid_model_folder` which start with a 2.

        Returns:
            pathlib.Path: Path to the most recent model file
        """

        model_folders = []

        # Loop over all subfolder of pid_model_folder and store the ones starting with
        # a 2 (date starts with 2000s)
        for folder in pid_model_folder.iterdir():
            if folder.name.startswith("2") is True and folder.is_dir() is True:
                model_folders.append(folder)

        # sort folders by date such that the most recent date is the first element
        for folder in sorted(model_folders, reverse=True):
            model_file = folder / self.MODEL_FILENAME
            if model_file.is_file() is True:
                model_folder = folder
                break
        # we did not find a valid model file
        else:
            msg = f"No model file found at save location: '{pid_model_folder}'"
            self.logger.error(msg)
            raise FileNotFoundError(msg)

        return model_file, model_folder

    def build_save_folder_path(self, pid):

        pid_model_folder = self.build_pid_model_folder_path(pid=pid)

        datetime_str = datetime.utcnow().strftime(self.FOLDER_DATETIME_FORMAT)

        save_folder = pid_model_folder / datetime_str

        return save_folder

    @classmethod
    def determine_model_age_from_path(cls, model_location):
        if model_location is None:
            return float("inf")

        # trained_models/<pid>/<YYYYMMDDHHMMSS>/model.bin
        datetime_string = model_location.parent.name
        model_datetime = datetime.strptime(datetime_string, cls.FOLDER_DATETIME_FORMAT)
        model_age = datetime.utcnow() - model_datetime

        model_age_days = model_age.days

        return model_age_days



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from ktpbase.log import logging

from openstf.model.prediction.xgboost.model.quantile import XGBQuantileModel
from openstf.model.serializer.xgboost.xgboost import XGBModelSerializer


class XGBQuantileModelSerializer(XGBModelSerializer):
    def __init__(self):
        super().__init__()
        self.logger = logging.get_logger(self.__class__.__name__)
        self.MODEL_FILENAME = "model_quantile.bin"

    # NOTE from general.py
    def load(self, pid):
        """Load the most recent model

        Args:
            pid ([type]): [description]
            forecast_type ([type]): [description]

        Raises:
            FileNotFoundError: If no recent model file was found

        Returns:
            [type]: [description]
        """

        pid_model_folder = self.build_pid_model_folder_path(pid=pid)

        try:
            model_file, model_folder = self.find_most_recent_model_file(
                pid_model_folder
            )
        except FileNotFoundError as e:
            self.logger.error(f"Can't load model, no recent model file found. {e}")
            return None, None

        self.logger.info(f"Loaded model from: {model_file}", model_file=model_file)

        xgb_quantile_model = XGBQuantileModel(model_file=model_file)

        xgb_quantile_model = self._add_atributes(model_folder, xgb_quantile_model)

        return xgb_quantile_model, model_file



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from pathlib import Path

PROJECT_ROOT = Path(__file__).parent.parent.absolute()



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pandas as pd


def apply_resample(df, to_resample, timedelta):
    """
    This functions return a new resampled dataframe with timedelta resolution

    Args:
        df (pandas.DataFrame): pandas TimeSeries DataFrame
        to_resample (dict): Dictionary that indicates what columns to resample and how.
            Example: {'column': ['mean'], ['sum']}
        timedelta (str): Resolution to resample to ('H', 'D', 'M', etc)
    Return:
        pandas.DataFrame: Resampled dataframe.
    """
    # Resampled dataframe
    resampled_data = pd.DataFrame()

    # Resample data for each column
    for column, resample_methods in to_resample.items():
        for method in resample_methods:
            new_column_name = f"{column}_{method}"
            resampled_data[new_column_name] = df[column].resample(timedelta).agg(method)

    return resampled_data


def apply_calender_features(df):
    # compute calender features
    df["is_monday"] = df.index.weekday == 0
    df["is_tuesday"] = df.index.weekday == 1
    df["is_wednesday"] = df.index.weekday == 2
    df["is_thurday"] = df.index.weekday == 3
    df["is_friday"] = df.index.weekday == 4
    df["is_saturday"] = df.index.weekday == 5
    df["is_sunday"] = df.index.weekday == 6
    df["is_weekday"] = df.index.weekday < 5
    df["is_weekendday"] = (df.index.weekday // 5) == 1
    df["month"] = df.index.month
    df["quarter"] = df.index.quarter

    return df


def apply_lag_features(df, lag_features, lag_times):
    lags = {}
    for t in lag_times:
        lags[t] = []
        for lag_feature in lag_features:
            # lag feature name
            name = lag_feature + "-" + str(t)
            # compute lag
            df[name] = df[lag_feature].shift(t)
            # ...
            lags[t].append(name)

    return df, lags


def apply_horizons(df, y_hor, lags, resample=False):
    # new dataframes
    dfs = []

    # copy of original dataframe
    df_copy = df.copy()

    for h in y_hor:
        # make copy of dataframe copy
        df_copy = df_copy.copy()

        # insert new column with horizon name
        df_copy["horizon"] = str(h)

        # alter copy
        for t in sorted(list(lags.keys())):
            # if horizon beyond lag variable
            if h >= t:
                # 'forget' lag features at time t
                for lag_feature in lags[t]:
                    df_copy[lag_feature] = np.nan

        # save copy
        dfs.append(df_copy)

    # concat dataframes
    df = pd.concat(dfs)

    if resample:
        # resample back to its original size
        # so that for each datapoint, one horizon variant exists
        i = np.array(range(len(df))).reshape(len(y_hor), -1).T
        c = np.random.choice(range(len(y_hor)), len(i), p=[0.8, 0.1, 0.1])
        i = i[range(len(i)), c]
        df = df.iloc[i]

    return df


def apply_outlier_removal(df, col):
    # compute high and low quantiles
    high_q = df[col].quantile(0.99)
    low_q = df[col].quantile(0.01)

    # remove outliers
    df = df[df[col] < high_q]
    df = df[df[col] > low_q]

    return df


def compute_classes(df):
    # compute quantiles
    quantiles = df.quantile([0.5, 0.75, 0.9]).values.flatten()

    # define class boundaries
    bounds = [0] + list(quantiles) + [df.max() + 0.1]

    # define classes
    classes = {}
    for i in range(len(bounds) - 1):
        classes[chr(65 + i)] = (bounds[i], bounds[i + 1])

    return classes


def apply_classes(df, y_col, classes):
    # convert y_col values to class labels
    y = df.copy()[y_col]
    for c in classes:
        lower, upper = classes[c]
        y[(df[y_col] >= lower) & (df[y_col] < upper)] = c

    # update original dataframe
    df[y_col] = y

    # TODO make this more general

    # Only use strings
    df = df.loc[df[y_col].apply(lambda x: isinstance(x, (str)) or np.isnan(x))]

    return df


def apply_capacity_features(
    df,
    y_col,
    y_hor,
    apply_class_labels=True,
    outlier_removal=False,
    load_profile_names=None,
):
    if load_profile_names is None:
        load_profile_names = []

    # resample load column
    load_resample_config = {"load": ["mean", "min", "max"]}
    load_df = apply_resample(df, load_resample_config, timedelta="D")

    # resample TDCV (Typical Domestic Consumption Values) load profiles
    load_profile_resample_config = {}
    for name in load_profile_names:
        load_profile_resample_config[name] = ["max"]
    load_profiles_df = apply_resample(df, load_profile_resample_config, timedelta="D")

    df = pd.concat([load_df, load_profiles_df], axis=1)

    # remove outliers
    if outlier_removal:
        df = apply_outlier_removal(df, col=y_col)

    # apply lag features
    lag_features = df.columns
    lag_times = list(range(1, 28))
    df, lags = apply_lag_features(df, lag_features, lag_times)

    # apply horizons (size of new data frame is: samples * horizons)
    df = apply_horizons(df, y_hor, lags, resample=False)

    # apply calender features
    df = apply_calender_features(df)

    # apply classes
    classes = None
    if apply_class_labels:
        classes = compute_classes(df[y_col])
        df = apply_classes(df, y_col=y_col, classes=classes)

    # drop lag features
    df = df.drop(columns=[f for f in lag_features if f != y_col])

    # sort index
    df = df.sort_index()

    return df, classes
